# TODO: video support

import sys
import time
import logging

import click
import multiprocessing
from PIL import Image, ImageSequence, UnidentifiedImageError
from moviepy import VideoFileClip
from tqdm import tqdm

from koba import __version__
from koba.core import core


multiprocessing.set_start_method("spawn")

logging.basicConfig(
    format="{asctime} - {levelname} - {message}",
    style="{",
    datefmt="%Y-%m-%d %H:%M",
    level=logging.ERROR
)

@click.command("koba")
@click.version_option(__version__)
@click.argument(
    "file",
    type=click.Path(
        exists=True,
        file_okay=True,
        readable=True,
    )
)
@click.option(
    "--char-aspect", default=2.0, show_default=True,
    help="Character height-to-width ratio (for aspect-correct output)."
)
@click.option(
    "--logging-level",
    default="ERROR",
    show_default=True,
    help="Set the logging verbosity. Options: CRITICAL, ERROR, WARNING, INFO, DEBUG."
)
@click.option(
    "--save-blocks", is_flag=True,
    help="Save each image block as a PNG file in the 'blocks/' directory."
)
@click.option(
    "--save-chars", is_flag=True,
    help="Save rendered character images in the 'chars/' directory."
)
@click.option(
    "--engine", "-e", default="diff", show_default=True,
    help="Similarity metric to use: brightness, ssim, diff, mse, ncc, hist, or cosine."
)
@click.option(
    "--font", default=None,
    help="Path to a custom TTF font file (overrides the default font)."
)
@click.option(
    "--char-range", default="32-126", show_default=True,
    help="Unicode range of characters to use, as start-end (e.g., 32-126)."
)
@click.option(
    "--stretch-contrast",
    is_flag=True,
    help="Stretch image contrast to use the full brightness range."
)
@click.option(
    "--scale",
    default=1.0,
    help="Sets the scale at which to render the image. Will be overwritten if image needs to be scaled down to allow proper processing."
)
@click.option(
    "--invert",
    is_flag=True,
    help="Inverts the image before processing."
)
@click.option(
    "--single-threaded",
    is_flag=True,
    help="Runs the whole program single-threaded."
)
@click.option(
    "--color",
    is_flag=True,
    help="Renders the image in color."
)
@click.option(
    "--fast-color",
    is_flag=True,
    help="Enables color and uses █ (U+2588) to improve processing speed. Only recommended for animated pictures."
)
def main(file, char_aspect, logging_level, save_blocks, save_chars, engine, font, char_range, stretch_contrast, scale, invert, single_threaded, color, fast_color):
    # update logging level
    logging.getLogger().setLevel(getattr(logging, logging_level.upper(), logging.ERROR))
    
    if fast_color:
        color = True
        char_range = "9608-9608"
    
    if font and not font.endswith(".ttf"):
        logging.critical("Please provide a font in TTF format.")
        sys.exit(1)

    split_range = char_range.split("-")
    start_char = split_range[0]
    end_char = split_range[1]
    if start_char.isnumeric() and end_char.isnumeric():
        start_char = int(start_char)
        end_char = int(end_char)
    else:
        logging.critical("Wrong format for char-range.")
        sys.exit(1)
    
    media_type = None
    
    # loading file and reading basic info
    try:
        img = Image.open(file)
        frame_count = getattr(img, 'n_frames', 1)
        if frame_count == 1:
            media_type = "image"
        else:
            media_type = "gif"
        frames = [frame.copy() for frame in ImageSequence.Iterator(img)]
    except UnidentifiedImageError:
        try:
            clip = VideoFileClip(file)
            frames = [Image.fromarray(f) for f in clip.iter_frames()]
            frame_count = len(frames)
            if frame_count > 1:
                media_type = "video"
            else:
                media_type = "image"
        except Exception as e:
            logging.critical(f"Unsupported or unreadable image/video format for file: {file}. Error: {e}")
            sys.exit(1)
    else:
        if not frames:
            logging.critical(f"No frames found in image file: {file}.")
            sys.exit(1)
    
    logging.info(f"Image has {frame_count} frame(s).")
    is_animated = frame_count > 1

    if is_animated:
        from koba.core import unify, charsets
        logging.info("Pre-rendering characters...")
        first_frame = frames[0]
        width, height = first_frame.size
        block_widths, block_heights, _ = core.calculate_block_sizes(width, height, char_aspect, scale)
        unique_shapes = {(w, h) for w in set(block_widths) for h in set(block_heights)}
        
        characters = charsets.get_range(start_char, end_char)
        
        if font:
            unify.font_path = font
        
        if abs(start_char - end_char) >= 400:
            for char in tqdm(characters, total=len(characters), desc="Loading fonts"):
                unify.get_font(char)

        with tqdm(total=len(unique_shapes) * len(characters), desc="Pre-rendering characters", disable=logging_level != "DEBUG") as pbar:
            unify.pre_render_characters(characters, unique_shapes, save_chars, pbar.update)
    
    frame_delays = []
    all_frames = []
    for i, frame in tqdm(enumerate(frames), total=frame_count, desc="Processing frames", disable=not is_animated):
        all_frames.append(core.process(
            frame, char_aspect, scale, engine, color, invert, stretch_contrast,
            save_blocks, start_char, end_char, save_chars, font, single_threaded, show_progress=not is_animated
        ))
        delay = 0
        if media_type == "gif":
            delay = frame.info.get("duration", 100) / 1000
        elif media_type == "video":
            delay = 1 / clip.fps
        elif media_type == "image":
            delay = 0.1
        
        if not delay or delay == 0:
            delay = 0.1
        frame_delays.append(delay)

    logging.debug(f"Frame delays: {frame_delays[:3]} ...")

    if media_type == "video":
        input("Press [Enter] to start playback: ")

    prev_lines = 0
    if is_animated:
        while True:
            for frame, delay in zip(all_frames, frame_delays):
                lines = frame.count('\n')
                if prev_lines > 0:
                    sys.stdout.write(f"\r\033[{prev_lines}A")
                    sys.stdout.write("\033[J")
                print(frame, end="")
                sys.stdout.flush()
                prev_lines = lines
                time.sleep(delay)
                
            if media_type == "video":
                try:
                    input("\nPress [Enter] to replay, or Ctrl+C to quit: ")
                    sys.stdout.write(f"\r\033[{prev_lines}A")
                    sys.stdout.write("\033[J")
                    sys.stdout.flush()
                except KeyboardInterrupt:
                    print("\nExiting.")
                    break
    else:
        print(all_frames[0])



import numpy as np
import pytest
from koba.core import unify


ENGINES = ["brightness", "ssim", "diff", "mse", "ncc", "hist", "cosine"]

# compare identical arrays
@pytest.mark.parametrize("engine", ENGINES)
def test_compare_perfect_match(engine, mocker):
    block = np.full((10, 10), 128, dtype=np.uint8)

    mocker.patch('koba.core.unify.get_char', return_value=block)

    similarity = unify.compare_character('a', block, False, engine)

    if engine == 'hist':
        # hist produces a larger error than the other engines
        assert similarity == pytest.approx(1.0, abs=1e-2)
    else:
        assert similarity == pytest.approx(1.0)

# compare black block to white block
@pytest.mark.parametrize("engine", ENGINES)
def test_compare_perfect_mismatch(engine, mocker):

    black_block = np.zeros((10, 10), dtype=np.uint8)
    white_block = np.full((10, 10), 255, dtype=np.uint8)

    mocker.patch('koba.core.unify.get_char', return_value=white_block)

    similarity = unify.compare_character('a', black_block, False, engine)

    if engine == 'ssim':
        # ssim can have a small error
        assert similarity == pytest.approx(0.0, abs=1e-4)
    else:
        assert similarity == pytest.approx(0.0)

def test_get_font():
    font = unify.get_font("a")
    if font:
        assert True
    else:
        assert False


import os
import sys
import math
import logging
import concurrent.futures

import numpy as np
from tqdm import tqdm
from PIL import Image, ImageOps

from koba.core import charsets, unify

def process_block(args):
    block, characters, engine, save_chars = args
    return unify.get_character(block, characters, save_chars, engine)

def calculate_block_sizes(width, height, char_aspect, scale):
    terminal_width = os.get_terminal_size().columns
    chars_width = terminal_width
    
    min_block_width = 10 / char_aspect
    min_block_height = 10

    scale = min(scale, 1.0)
    
    max_chars_width_by_block_width = int(width // min_block_width)
    max_chars_width_by_block_height = int(height // min_block_height * char_aspect)

    max_chars_width = min(max_chars_width_by_block_width, max_chars_width_by_block_height)

    chars_width = min(int(terminal_width * scale), terminal_width, max_chars_width)

    chars_width = max(chars_width, 1)

    chars_height = math.ceil((height * chars_width / width) / char_aspect)
    
    block_widths = [width // chars_width] * chars_width
    for i in range(width % chars_width):
        block_widths[i] += 1

    block_heights = [height // chars_height] * chars_height
    for i in range(height % chars_height):
        block_heights[i] += 1
        
    return block_widths, block_heights, chars_width

def process(img, char_aspect, scale, engine, color, invert, stretch_contrast, save_blocks, start_char, end_char, save_chars, font, single_threaded, show_progress=False):
    if color:
        img_color = img.convert("RGB")
        img_arr_color = np.array(img_color)
    
    img = img.convert("L")
    if invert:
        img = ImageOps.invert(img)
    if stretch_contrast:
        img = ImageOps.autocontrast(img)
    
    img_arr = np.array(img)
    height, width = img_arr.shape[:2]
    logging.debug(f"Image is {width}x{height} pixels.")
    
    block_widths, block_heights, chars_width = calculate_block_sizes(width, height, char_aspect, scale)

    logging.debug(f"Image will be {chars_width}x{len(block_heights)} chars.")
    logging.debug(f"Block widths: {block_widths[:5]}... (total {len(block_widths)})")
    logging.debug(f"Block heights: {block_heights[:5]}... (total {len(block_heights)})")

    if min(block_heights + block_widths) <= 7 and str(engine).lower() == "ssim":
        logging.critical("Image blocks are too small for SSIM. Please use another engine.")
        sys.exit(1)

    # splitting into blocks with variable sizes
    block_colors = []
    blocks = []
    y = 0
    for bh in block_heights:
        x = 0
        for bw in block_widths:
            block = img_arr[y:y+bh, x:x+bw]                
            blocks.append(block)
            x += bw
        y += bh
    
    if color:
        y = 0
        for bh in block_heights:
            x = 0
            for bw in block_widths:
                block = img_arr_color[y:y+bh, x:x+bw]
                block_color = block.mean(axis=(0, 1)).astype(int)            
                block_colors.append(block_color)
                x += bw
            y += bh
    
    if color:
        if len(block_colors) != len(blocks):
            logging.critical("Block colors don't match the blocks. Falling back to grayscale.")
            color = False
        else:
            logging.debug(f"Block colors: {block_colors[:3]} ...")
    
    logging.debug(f"{len(blocks)} blocks created.")
    
    if save_blocks:
        logging.info("Saving image blocks...")
        blocks_dir = "blocks"
        os.makedirs(blocks_dir, exist_ok=True)
        for idx, block in enumerate(blocks):
            img_block = Image.fromarray(block)
            img_block.save(os.path.join(blocks_dir, f"{idx:04d}.png"))
    
    print_chars = ""
    
    if start_char == end_char:
        all_chars = chr(start_char) * len(blocks)
    else:
        characters = charsets.get_range(start_char, end_char)
        all_chars = ""

        # prepare arguments
        unique_blocks = {block.tobytes(): block for block in blocks}
        unique_block_args = [(block, characters, engine.lower(), save_chars) for block in unique_blocks.values()]

        if font:
            unify.font_path = font

        block_results = {}
        if not single_threaded:
            with concurrent.futures.ProcessPoolExecutor() as executor:
                future_to_block = {
                    executor.submit(unify.get_character, *arg): arg[0] for arg in unique_block_args
                }
                try:
                    for future in tqdm(
                        concurrent.futures.as_completed(future_to_block),
                        total=len(future_to_block),
                        desc="Processing unique blocks",
                        disable=not show_progress
                    ):
                        block = future_to_block[future]
                        block_results[block.tobytes()] = future.result()
                except ValueError as e:
                    logging.critical(str(e))
                    sys.exit(1)
        else:
            for args in tqdm(
                unique_block_args,
                desc="Processing unique blocks (single-threaded)",
                disable=not show_progress
            ):
                block, characters, engine, save_chars = args
                result = unify.get_character(block, characters, save_chars, engine)
                block_results[block.tobytes()] = result

        results = [block_results[block.tobytes()] for block in blocks]
        
        all_chars = "".join(results)
    
    lines = [all_chars[i:i+chars_width] for i in range(0, len(all_chars), chars_width)]
    flat_chars = "".join(lines)
    logging.debug(f"Char count: {len(flat_chars)}")
    logging.debug(f"Colored blocks: {len(block_colors)}")

    color_idx = 0
    for c in '\n'.join(lines):
        if c == '\n':
            print_chars += '\n'
        elif color_idx < len(block_colors):
            r, g, b = block_colors[color_idx]
            print_chars += f"\033[38;2;{r};{g};{b}m{c}\033[0m"
            color_idx += 1
        else:
            print_chars += c
            
    return print_chars
